package org.processmining.localprocessmodelconformance.plugins;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.lpm.util.LocalProcessModel;
import org.processmining.lpm.util.LocalProcessModelRanking;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetImpl;
import org.processmining.models.semantics.petrinet.Marking;

@Plugin(name = "From CloGSgrow patterns to LPM Ranking", parameterLabels = { "Closed Repetitive Gapped Sequential Patterns" }, returnLabels = { "Local Process Model Ranking" }, returnTypes = { LocalProcessModelRanking.class }, userAccessible = true, categories = { PluginCategory.Discovery })
public class FromPatternSetToLocalProcessModelRanking {
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "From CloGSgrow patterns to LPM Ranking", requiredParameterLabels = { 0 })
	public LocalProcessModelRanking run(PluginContext context, XLog log) {
		LocalProcessModelRanking ranking = new LocalProcessModelRanking();
		for(XTrace trace : log){
			ranking.addNet(traceToNet(trace));
		}
		return ranking;
	}

	public static LocalProcessModel traceToNet(XTrace trace) {
		Petrinet net = new PetrinetImpl("net");
		Place sourcePlace = net.addPlace("source");
		Place lastPlace = sourcePlace;
		int placeCounter = 1;
		int i=1;
		for(XEvent event : trace){
			Transition t = net.addTransition(getActivity(event));
			net.addArc(lastPlace, t);
			Place newPlace = net.addPlace("place "+placeCounter++);
			net.addArc(t, newPlace);
			if(getMergeOperation(event).equals("Length-1-Loop")){
				Transition s = net.addTransition("silent");
				s.setInvisible(true);
				net.addArc(newPlace, s);
				net.addArc(s, lastPlace);
				if(i==trace.size()){ // make unique sink place
					Transition s2 = net.addTransition("silent");
					s2.setInvisible(true);
					net.addArc(newPlace, s2);
					newPlace = net.addPlace("place "+placeCounter++);
					net.addArc(s2, newPlace);
				}
			}	
			lastPlace = newPlace;
			i++;
		}
		return new LocalProcessModel(makeAcceptingNet(net, sourcePlace, lastPlace));
	}
	
	public static AcceptingPetriNet makeAcceptingNet(Petrinet net, Place initialPlace, Place finalPlace){
		Marking initialMarking = new Marking();
		initialMarking.add(initialPlace);
		Marking finalMarking = new Marking();
		finalMarking.add(finalPlace);
		return new AcceptingPetriNetImpl(net, initialMarking, finalMarking);
	}
	
	public static String getMergeOperation(XEvent event){
		String mergeOperation = "NotMerged";
		if(event.getAttributes().containsKey("MergeOperation"))
			mergeOperation = event.getAttributes().get("MergeOperation").toString();
		return mergeOperation;
	}
	
	public static String getActivity(XEvent event){
		return event.getAttributes().get("concept:name").toString();
	}
}
