package org.processmining.localprocessmodelconformance.plugins;

import java.util.HashSet;
import java.util.Set;

import org.deckfour.xes.model.XLog;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.logenhancement.abstraction.PatternStructureException;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionVisitorException;
import org.processmining.logenhancement.abstraction.model.syntax.ParseException;
import org.processmining.lpm.util.LocalProcessModel;
import org.processmining.lpm.util.LocalProcessModelRanking;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;

@Plugin(name = "Select Local Process Models Greedily (Revisited)", parameterLabels = { "Log",
"Local Process Models" }, returnLabels = { "Local Process Models" }, returnTypes = {
		LocalProcessModelRanking.class }, userAccessible = true, categories = { PluginCategory.Enhancement })
public class GreedyLPMSelectorRevisited {
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Select Local Process Models Greedily (Revisited)", requiredParameterLabels = { 0, 1 })
	public LocalProcessModelRanking greedySelector(PluginContext context, XLog log, LocalProcessModelRanking ranking) throws ControlFlowAlignmentException, DataAlignmentException, PatternStructureException, ParseException, CompositionVisitorException {
		Petrinet[] nonAcceptingNets = new Petrinet[ranking.getSize()];
		for(int i=0; i<ranking.getSize(); i++){
			nonAcceptingNets[i] = ranking.getNet(i).getAcceptingPetriNet().getNet();
		}
		return greedySelector(log, nonAcceptingNets, new HashSet<Petrinet>(), 0);
	}

	private static LocalProcessModelRanking greedySelector(XLog log, Petrinet[] nets, Set<Petrinet> selectedNets, double bestFscore) throws ControlFlowAlignmentException, DataAlignmentException, PatternStructureException, ParseException, CompositionVisitorException {
		Set<Petrinet> bestLocalNets = new HashSet<Petrinet>();
		double localBestFscore = 0;
		for(Petrinet newCandidateNet : nets){
			if(selectedNets.contains(newCandidateNet))
				continue;
			Petrinet[] netsToBeTested = new Petrinet[selectedNets.size()+1];
			int i=0;
			for(Petrinet existingNet : selectedNets){
				netsToBeTested[i] = existingNet;
				i++;
			}
			netsToBeTested[i] = newCandidateNet;
			Object[] result = RetrieveLPMEpisodesForLPMCollection.matchPatternsToLogEpisodes(log, netsToBeTested);
			double coverage = (double) result[4];
			double precision = (double) result[5];
			double fscore = (2*coverage*precision)/(coverage+precision);
			if(fscore>localBestFscore){
				localBestFscore = fscore;
				bestLocalNets = new HashSet<Petrinet>();
				for(Petrinet pn : netsToBeTested)
					bestLocalNets.add(pn);
			}
		}
		if(localBestFscore>bestFscore){
			return greedySelector(log, nets, bestLocalNets, localBestFscore);
		}else{
			LocalProcessModelRanking ranking = new LocalProcessModelRanking();
			for(Petrinet pn : selectedNets)
				ranking.addNet(new LocalProcessModel(new AcceptingPetriNetImpl(pn)));
			return ranking;
		}
	}
}