package org.processmining.localprocessmodelconformance.plugins;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeDiscreteImpl;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;

@Plugin(name = "Merge Loop", parameterLabels = { "Log" }, returnLabels = { "Log with Loops Merged" }, returnTypes = { XLog.class }, userAccessible = true, categories = { PluginCategory.Enhancement })
public class MergeLoop {
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Merge Loop", requiredParameterLabels = { 0 })
	public XLog run(PluginContext context, XLog log) {
		XLog clone = (XLog) log.clone();
		for(XTrace trace : clone){
			for(int n=1; n<trace.size()/2; n++){
				trace = searchAndReplaceLengthNLoop(trace, n);
			}
		}
		collapseIdenticalTraces(clone);
		return clone;
	}
	
	private void collapseIdenticalTraces(XLog log){
		Set<XTrace> toRemoveSet = new HashSet<XTrace>();
		for(int i=0; i<log.size(); i++){
			for(int j=(i+1); j<log.size(); j++){
				if(getActivityRepresentation(log.get(i)).equals(getActivityRepresentation(log.get(j)))){
					if(getMerged(log.get(j))){
						toRemoveSet.add(log.get(i));
						setSupport(log.get(j), getSupport(log.get(i))+getSupport(log.get(j)));
					}else{
						toRemoveSet.add(log.get(j));
						setSupport(log.get(i), getSupport(log.get(i))+getSupport(log.get(j)));
					}
				}
			}
		}
		for(XTrace t : toRemoveSet)
			log.remove(t);
		toRemoveSet.clear();
		for(XTrace t: log)
			if(t.size()==1)
				toRemoveSet.add(t);
		for(XTrace t : toRemoveSet)
			log.remove(t);
	}
	
	private List<String> getActivityRepresentation(XTrace trace) {
		List<String> result = new ArrayList<String>();
		for(XEvent e : trace)
			result.add(getActivity(e));
		return result;
	}

	private XTrace searchAndReplaceLengthNLoop(XTrace traceClone, int n) {
		boolean patternFound = true;
		while(patternFound){
			Deque<String> lastNA = new ArrayDeque<String>(n);
			Deque<XEvent> lastNE = new ArrayDeque<XEvent>(n);
			Deque<String> compareNA = new ArrayDeque<String>(n);
			Deque<XEvent> compareNE = new ArrayDeque<XEvent>(n);
			Set<XEvent> toRemove = new HashSet<XEvent>();
			patternFound = false;
			for(XEvent e : traceClone){
				if(lastNA.size()<n){
					lastNA.add(getActivity(e));
					lastNE.add(e);
				}else{
					if(compareNA.size()>=n){
						compareNA.poll();
						compareNE.poll();
					}
					compareNA.add(lastNA.poll());
					compareNE.add(lastNE.poll());
					lastNA.add(getActivity(e));
					lastNE.add(e);
					if(new ArrayList<String>(compareNA).equals(new ArrayList<String>(lastNA))){ // ArrayDeque (surprisingly) doesn't implement equals, so make lists first!
						// found instance!
						toRemove.addAll(lastNE);
						patternFound = true;
						compareNE.getFirst().getAttributes().put("MergeOperation", new XAttributeLiteralImpl("MergeOperation", "Length-"+n+"-Loop"));
						setMerged(traceClone);
						lastNA.clear();
						lastNE.clear();
						compareNA.clear();
						compareNE.clear();
					}
				}
			}
			traceClone.removeAll(toRemove);
		}
		return traceClone;
	}

	public static String getActivity(XEvent event){
		return event.getAttributes().get("concept:name").toString();
	}
	
	public static boolean getMerged(XTrace trace){
		return trace.getAttributes().containsKey("merged");
	}
	
	public static void setMerged(XTrace trace){
		trace.getAttributes().put("merged", new XAttributeLiteralImpl("merged", "TRUE"));
	}
	
	public static void setSupport(XTrace trace, int value){
		trace.getAttributes().put("support", new XAttributeDiscreteImpl("support", value));
	}
	
	public static int getSupport(XTrace trace){
		return (int) ((XAttributeDiscrete) trace.getAttributes().get("support")).getValue();
	}
	
}
